/* as.c - The NEW Artificial Stupidity program...  Completely table
		  configurable (for adding/changing personalities)...  Portable
		  version (with my own versions of WS library functions used)...

	By: Robert Seace
		120 Brainerd Road, Apt. 2
		Allston, MA  02134
		(617) 277-8835
		ras@world.std.com

	Public Domain code...  Do with it what you will...  But, PLEASE
	make note of any changes, and increment the version#, if you are
	re-distributing it...

	V1.00 - Baseline...
		[Rob Seace, 1-Nov-94]
*/

#include "as.h"

#ifdef __USAGE
V1.00 - PORTABLE VERSION

AS - the table-driven Artificial Stupidity program

%C [-p <file>] [-e]

-p <file>    Personality table file to use. (Default = "as.tbl")
-e           Echo the input that is read.  (Useful for re-directed, or
             piped stdin.)

#endif

void prt_usage_msg (progname)
	string	progname;				/* name of the program					*/
{
	fprintf (stderr, "V1.00 - PORTABLE VERSION\n\n");

	fprintf (stderr, "AS - the table-driven Artificial Stupidity program\n\n");

	fprintf (stderr, "%s [-p <file>] [-e]\n\n", progname);

	fprintf (stderr, "-p <file>    Personality table file to use. (Default = '%s')\n", AS_DFLT_PTBL_FILE);
	fprintf (stderr, "-e           Echo the input that is read.  (Useful for re-directed, or\n");
	fprintf (stderr, "             piped stdin.)\n\n");
}


struct as_config_struct	AS_Configs;			/* program configurations	*/
as_node					AS_Head;			/* head of list				*/


void main (argc, argv)
	int		argc;
	char	*argv[];
{
	int		opt;						/* returned by getopt()				*/
	bool	echo_input;					/* TRUE => echo the read-in input	*/
	string	ptbl_file;					/* name of personality table file	*/
	char	buf[AS_LINE_LEN+1];			/* line input buffer				*/

	ptbl_file = AS_DFLT_PTBL_FILE;
	echo_input = FALSE;

#ifdef YOU_HAVE_GETOPT

	while ((opt = getopt (argc, argv, "ep:")) ISNT -1) {
		switch (opt) {
			case	'e':
				echo_input = TRUE;
				break;
			case	'p':
				ptbl_file = optarg;
				break;
			default:
				prt_usage_msg (argv[0]);
				exit (ERROR);
		}
	}

#else	/* YOU_HAVE_GETOPT ... not! */

	for (opt = 0; opt < argc; opt++) {
		if (streq (argv[opt], "-e")) {
			echo_input = TRUE;
		} else if (streq (argv[opt], "-p")) {
			ptbl_file = argv[opt+1];
			opt++;	/* must skip ahead TWO parameters */
		} else {
			fprintf (stderr, "Unknown command line parameter (%s).\n\n", argv[opt]);
			prt_usage_msg (argv[0]);
			exit (ERROR);
		}
	}

#endif	/* YOU_HAVE_GETOPT */

	as_ptbl_init (ptbl_file);

#ifdef DEBUGGING_MODE
	/* debugging info */
	as_debug_print_list (AS_Head, 1);
	as_debug_print_strlist (AS_Configs.people_names, 1);
	as_debug_print_strlist (AS_Configs.quit_strs, 1);
	exit (ERROR);	/* debugging info only */
#endif

	as_print_greeting ();

	while (as_read_input (buf, echo_input)) {
		as_process_input (buf);
	}

	as_print_parting ();

	as_ptbl_free ();

	exit (SUCCESS);
}

