/* as_util.c - Utility functions for AS...
*/

#include "as.h"


/* as_print_greeting - Prints main initial greeting on start-up...
*/

void as_print_greeting ()
{
	if (AS_Configs.greeting_str)
		fprintf (stdout, "%s", AS_Configs.greeting_str);
}

/* as_print_parting - Prints main parting message on leaving...
*/

void as_print_parting ()
{
	if (AS_Configs.parting_str)
		fprintf (stdout, "%s", AS_Configs.parting_str);
}


/* as_read_input - Prints prompt, if any, reads input from stdin, filling
					passed buffer.  Returns TRUE if read successful, and
					not a quit string, else FALSE to quit.
*/

bool as_read_input (buf, echo_input)
	string	buf;							/* input buffer to fill			*/
	bool	echo_input;						/* TRUE => echo what was read	*/
{
	for (;;) {
		if (AS_Configs.prompt_str)
			fprintf (stdout, "%s ", AS_Configs.prompt_str);

		if (fgets (buf, AS_LINE_LEN, stdin) IS NULL)	/* EOF */
			return (FALSE);		/* always quit on EOF */

		if (echo_input)
			fputs (buf, stdout);	/* write it as read */

		strupper (strtrim (buf));

		if (buf[0])		/* non-null input */
			break;
	}

	if (as_str_in_strlist (buf, AS_Configs.quit_strs, TRUE)) {
		return (FALSE);		/* user wants to quit */
	} else {
		return (TRUE);		/* valid input; buf ready to process */
	}
}

/* as_process_input - Breaks up read-in line into component tokens,
					  checking for each one in main personality list.  If
					  if finds one, it then begins anew searching for any
					  co-dependant strings.  If it finds one of those in
					  the previously found sub-list, it prints the output
					  defined as a response for that match (may be a random
					  string out of a list of many)...
					  If it fails to find any match at all in it's personality
					  table, it will not respond.  (Personality tables should
					  generally have a final default case, using "*" for both
					  the main and sub match string...)
*/

void as_process_input (buf)
	string	buf;							/* buffer containing input line	*/
{
	string	words[AS_MAX_WORDS+1];			/* array of tokens from line	*/
	as_node	n;								/* pointer to node in list		*/
	as_node	n2;								/* pointer to node in list		*/
	int		i;								/* loop counter					*/
	int		j;								/* loop counter					*/
static	char	delims[MY_BUF_LEN+1];		/* word delimiters				*/

	if (delims[0] IS '\0')
		sprintf (delims, "-.,?!;:() \t\n\r");

	str2array (buf, delims, TRUE, words);	/* break up line into words */

	if (NOT words[0])	/* no words in line */
		return;

	/* Don't even attempt to understand the following demented search	*/
	/* logic...  You'll just go mad...									*/

	for (i = 0; words[i]; i++) {
		if (n = as_str_in_list (words[i], AS_Head, FALSE)) {
			for (j = i+1; words[j]; j++) {
				if (n2 = as_str_in_list (words[j], n->str_list, FALSE)) {
					as_respond (n2->str_list, words[i], words[j]);
					return;		/* only respond once */
				}
			}
			for (j = 0; (j < i) AND words[j]; j++) {
				if (n2 = as_str_in_list (words[j], n->str_list, FALSE)) {
					as_respond (n2->str_list, words[i], words[j]);
					return;		/* only respond once */
				}
			}
		}
	}
	for (i = 0; words[i]; i++) {
		if (n = as_str_in_list (words[i], AS_Head, FALSE)) {
			for (j = i+1; words[j]; j++) {
				if (n2 = as_str_in_list (words[j], n->str_list, TRUE)) {
					as_respond (n2->str_list, words[i], words[j]);
					return;		/* only respond once */
				}
			}
			for (j = 0; (j < i) AND words[j]; j++) {
				if (n2 = as_str_in_list (words[j], n->str_list, TRUE)) {
					as_respond (n2->str_list, words[i], words[j]);
					return;		/* only respond once */
				}
			}
		}
	}
	for (i = 0; words[i]; i++) {
		if (n = as_str_in_list (words[i], AS_Head, TRUE)) {
			for (j = i+1; words[j]; j++) {
				if (n2 = as_str_in_list (words[j], n->str_list, FALSE)) {
					as_respond (n2->str_list, words[i], words[j]);
					return;		/* only respond once */
				}
			}
			for (j = 0; (j < i) AND words[j]; j++) {
				if (n2 = as_str_in_list (words[j], n->str_list, FALSE)) {
					as_respond (n2->str_list, words[i], words[j]);
					return;		/* only respond once */
				}
			}
		}
	}
	for (i = 0; words[i]; i++) {
		if (n = as_str_in_list (words[i], AS_Head, TRUE)) {
			for (j = i+1; words[j]; j++) {
				if (n2 = as_str_in_list (words[j], n->str_list, TRUE)) {
					as_respond (n2->str_list, words[i], words[j]);
					return;		/* only respond once */
				}
			}
			for (j = 0; (j < i) AND words[j]; j++) {
				if (n2 = as_str_in_list (words[j], n->str_list, TRUE)) {
					as_respond (n2->str_list, words[i], words[j]);
					return;		/* only respond once */
				}
			}
		}
	}

	for (i = 0; words[i]; i++) {
		if (n = as_str_in_list (words[i], AS_Head, FALSE)) {
			for (j = 0; words[j]; j++) {
				if (n2 = as_str_in_list (words[j], n->str_list, FALSE)) {
					as_respond (n2->str_list, words[i], words[j]);
					return;		/* only respond once */
				}
			}
		}
	}
	for (i = 0; words[i]; i++) {
		if (n = as_str_in_list (words[i], AS_Head, FALSE)) {
			for (j = 0; words[j]; j++) {
				if (n2 = as_str_in_list (words[j], n->str_list, TRUE)) {
					as_respond (n2->str_list, words[i], words[j]);
					return;		/* only respond once */
				}
			}
		}
	}
	for (i = 0; words[i]; i++) {
		if (n = as_str_in_list (words[i], AS_Head, TRUE)) {
			for (j = 0; words[j]; j++) {
				if (n2 = as_str_in_list (words[j], n->str_list, FALSE)) {
					as_respond (n2->str_list, words[i], words[j]);
					return;		/* only respond once */
				}
			}
		}
	}
	for (i = 0; words[i]; i++) {
		if (n = as_str_in_list (words[i], AS_Head, TRUE)) {
			for (j = 0; words[j]; j++) {
				if (n2 = as_str_in_list (words[j], n->str_list, TRUE)) {
					as_respond (n2->str_list, words[i], words[j]);
					return;		/* only respond once */
				}
			}
		}
	}

	/* if it didn't get caught by the above psychotic method, then find and
		use the default, if one exists... */

	i = 0;

	if (words[1])
		j = 1;
	else
		j = 0;

	if (n = as_list_find_default (AS_Head)) {
		if (n2 = as_list_find_default (n->str_list)) {
			as_respond (n2->str_list, words[i], words[j]);
			return;		/* only respond once */
		}
	}

	/* no default case!  Screw 'em...  No response... */

}

/* as_respond - Handles responding to user, based on passed response set...
*/

void as_respond (str_list, main_str, sub_str)
	as_node	str_list;					/* pointer to the response set		*/
	string	main_str;					/* main matched string				*/
	string	sub_str;					/* sub/secondary matched string		*/
{
	as_node	listp;						/* pointer to a node of the list	*/
	int		idx;						/* index into str list, if any		*/
	int		last_idx;					/* last index into str list, if any	*/
	char	buf[AS_LINE_LEN+1];			/* output buffer					*/

	last_idx = -1;

	for (listp = str_list; listp; listp = listp->next) {
		idx = as_format_output (buf, listp, main_str, sub_str, last_idx);

		fprintf (stdout, "%s: %s\n", as_char_name (listp->code), buf);

		last_idx = idx;
	}
}

/* as_format_output - Formats output string for a single response.  Returns
					  the index from last dependant list.
*/

int as_format_output (buf, listp, main_str, sub_str, last_idx)
	string	buf;						/* output buffer to fill			*/
	as_node	listp;						/* pointer to a node of the list	*/
	string	main_str;					/* main matched string				*/
	string	sub_str;					/* sub/secondary matched string		*/
	int		last_idx;					/* last index into str list, if any	*/
{
	int		i;							/* loop counter / string index		*/
	int		j;							/* loop counter / string index		*/
	int		len;						/* # of chars written by sprintf()	*/
	string	str;						/* response string (with macros)	*/
	char	tmp[AS_LINE_LEN+1];			/* temp buffer for recursion		*/

	if (listp IS NULL) {
		buf[0] = '\0';
		return (last_idx);
	}

	str = listp->str;

	if (str IS NULL) {
		buf[0] = '\0';
		return (last_idx);
	}

	for (i = 0, j = 0; str[i]; ) {
		if (str[i] IS '%') {		/* macro marker */
			switch (str[i+1]) {		/* check next char */
				case	'M':	/* main match string */
					len = sprintf (&(buf[j]), "%s", main_str);
					i += 2;
					j += len;
					break;
				case	'S':	/* secondary match string */
					len = sprintf (&(buf[j]), "%s", sub_str);
					i += 2;
					j += len;
					break;
				case	'{':	/* dependant list */
					if (last_idx < 0)
						last_idx = as_rand_idx (listp->str_list);

					/* recursion */
					as_format_output (tmp, as_node_at_idx (last_idx, listp->str_list), main_str, sub_str, last_idx);

					len = sprintf (&(buf[j]), "%s", tmp);
					i += 2;
					j += len;
					break;
				case	'[':	/* non-dependant list */
					/* recursion */
					as_format_output (tmp, as_node_at_idx (as_rand_idx (listp->str_list), listp->str_list), main_str, sub_str, last_idx);

					len = sprintf (&(buf[j]), "%s", tmp);
					i += 2;
					j += len;
					break;
				default:	/* either number or non-recogized macro */
					if (isdigit (str[i+1])) {	/* character number */
						len = sprintf (&(buf[j]), "%s", as_char_name (atoi (&(str[i+1]))));
						i++;
						while (isdigit (str[i]))
							i++;
						j += len;
					} else {	/* unrecognized macro; print as literal */
						buf[j] = str[i+1];
						i += 2;
						j++;
					}
					break;
			}
		} else {	/* normal, non-macro character */
			buf[j++] = str[i++];
		}
	}

	buf[j] = '\0';

	return (last_idx);
}

/* as_ptbl_init - Reads passed file, building in-memory personality table...
*/

void as_ptbl_init (fname)
	string	fname;						/* name of personality table file	*/
{
	file	fp;							/* file pointer for table file		*/
	char	buf[AS_LINE_LEN+1];			/* line-input buffer				*/
	char	tmp[AS_LINE_LEN+1];			/* temp buffer						*/
	char	tmp2[AS_LINE_LEN+1];		/* temp buffer						*/
	string	p, p2;						/* temp string pointers				*/
	as_str_node	n;						/* string node						*/

	AS_Head = NULL;
	AS_Configs.prompt_str = AS_Configs.greeting_str = AS_Configs.parting_str = NULL;
	AS_Configs.people_names = AS_Configs.quit_strs = NULL;

	srand ((unsigned)(time(NULL)*getpid()));

	if ((fp = fopen (fname, "r")) IS NULL) {
		fprintf (stderr, "Unable to open personality configuration table (%s).\n", fname);
		exit (ERROR);
	}

	while (fgets (buf, AS_LINE_LEN, fp)) {
		strtrim (buf);
		if (buf[0] IS '#')	/* comment */
			continue;	/* ignore it */

		if (buf[0] IS '!') {	/* main match string */
			if (sscanf (buf, "!%[^:]", tmp) ISNT 1) {
				fprintf (stderr, "Invalid line in personality table (%s).\nLine = %s\n", fname, buf);
				exit (ERROR);
			}

			strupper (strtrim (tmp));

			as_make_new_main_node (tmp, fp, fname);
		} else if (NOT strncmp (buf, "PEOPLE:", 7)) {	/* list of people */
			p = buf+7;
			while (p2 = strchr (p, ';')) {
				*p2 = '\0';		/* null terminate */

				if ((n = (as_str_node) calloc (1, AS_STR_NODE_LEN)) IS NULL) {
					fprintf (stderr, "Not enough memory for people name node.\n");
					exit (ERROR);
				}

				strcpy (tmp, p);
				strtrim (tmp);

				if ((n->str = (string) calloc (1, strlen (tmp) + 1)) IS NULL) {
					fprintf (stderr, "Not enough memory for people name string.\n");
					exit (ERROR);
				}

				strcpy (n->str, tmp);

				if (AS_Configs.people_names IS NULL) {
					n->next = n->prev = NULL;
					AS_Configs.people_names = n;
				} else {
					as_add_strnode (n, AS_Configs.people_names);
				}

				p = p2+1;
			}
		} else if (NOT strncmp (buf, "QUIT:", 5)) {		/* list of quit strs */
			p = buf+5;
			while (p2 = strchr (p, ';')) {
				*p2 = '\0';		/* null terminate */

				if ((n = (as_str_node) calloc (1, AS_STR_NODE_LEN)) IS NULL) {
					fprintf (stderr, "Not enough memory for quit string node.\n");
					exit (ERROR);
				}

				strcpy (tmp, p);
				strupper (strtrim (tmp));

				if ((n->str = (string) calloc (1, strlen (tmp) + 1)) IS NULL) {
					fprintf (stderr, "Not enough memory for quit string.\n");
					exit (ERROR);
				}

				strcpy (n->str, tmp);

				if (AS_Configs.quit_strs IS NULL) {
					n->next = n->prev = NULL;
					AS_Configs.quit_strs = n;
				} else {
					as_add_strnode (n, AS_Configs.quit_strs);
				}

				p = p2+1;
			}
		} else if (NOT strncmp (buf, "PROMPT:", 7)) {	/* prompt string */
			strcpy (tmp2, buf+7);
			strtrim (tmp2);
			as_sprintf (tmp, tmp2);
			if ((AS_Configs.prompt_str = (string) calloc (1, strlen (tmp) + 1)) IS NULL) {
				fprintf (stderr, "Not enough memory for prompt string.\n");
				exit (ERROR);
			}
			strcpy (AS_Configs.prompt_str, tmp);
		} else if (NOT strncmp (buf, "GREETING:", 9)) {	/* greeting string */
			strcpy (tmp2, buf+9);
			strtrim (tmp2);
			as_sprintf (tmp, tmp2);
			if ((AS_Configs.greeting_str = (string) calloc (1, strlen (tmp) + 1)) IS NULL) {
				fprintf (stderr, "Not enough memory for greeting string.\n");
				exit (ERROR);
			}
			strcpy (AS_Configs.greeting_str, tmp);
		} else if (NOT strncmp (buf, "PARTING:", 8)) {	/* parting string */
			strcpy (tmp2, buf+8);
			strtrim (tmp2);
			as_sprintf (tmp, tmp2);
			if ((AS_Configs.parting_str = (string) calloc (1, strlen (tmp) + 1)) IS NULL) {
				fprintf (stderr, "Not enough memory for parting string.\n");
				exit (ERROR);
			}
			strcpy (AS_Configs.parting_str, tmp);
		} else {	/* unknown */
			fprintf (stderr, "Invalid line in personality configuration table (%s).\nLine = %s\n", fname, buf);
			exit (ERROR);
		}
	}

	fclose (fp);
}

/* as_ptbl_free - Frees up all allocated memory in personality table...
*/

void as_ptbl_free ()
{
	as_list_free (AS_Head);
	as_strlist_free (AS_Configs.people_names);
	as_strlist_free (AS_Configs.quit_strs);

	AS_Head = NULL;
	AS_Configs.people_names = AS_Configs.quit_strs = NULL;
}


/* as_make_new_main_node - Reads and processes lines from passed file pointer
							for a new main node...
*/

void as_make_new_main_node (main_str, fp, fname)
	string	main_str;					/* main match string				*/
	file	fp;							/* file pointer to read from		*/
	string	fname;						/* name of personality table file	*/
{
	long	old_pos;					/* old saved position to fseek() to	*/
	char	buf[AS_LINE_LEN+1];			/* line-input buffer				*/
	char	tmp[AS_LINE_LEN+1];			/* temp buffer						*/
	char	num_buf[AS_LINE_LEN+1];		/* character number buffer			*/
	char	resp_buf[AS_LINE_LEN+1];	/* response buffer					*/
	as_node	n;							/* a new node						*/
	as_node	n2;							/* a new node						*/
	as_node	n3;							/* a new node						*/

	if ((n2 = (as_node) calloc (1, AS_NODE_LEN)) IS NULL) {
		fprintf (stderr, "Not enough memory for main node.\n");
		exit (ERROR);
	}

	if ((n2->str = (string) calloc (1, strlen (main_str) + 1)) IS NULL) {
		fprintf (stderr, "Not enough memory for response string.\n");
		exit (ERROR);
	}

	strcpy (n2->str, main_str);

	n2->str_list = NULL;

	if (AS_Head IS NULL) {
		n2->next = n2->prev = NULL;
		AS_Head = n2;
	} else {
		as_add_node (n2, AS_Head);
	}

	old_pos = ftell (fp);
	n3 = NULL;

	while (fgets (buf, AS_LINE_LEN, fp)) {
		strtrim (buf);
		if (buf[0] IS '#')	/* comment */
			continue;	/* ignore it */

		if (buf[0] IS '!') {	/* main match string */
			fseek (fp, old_pos, SEEK_SET);	/* backup a line */
			break;
		}

		if (sscanf (buf, "%[^:]:%[^\n]", num_buf, resp_buf) IS 2) {	/* current sub-string */
			strtrim (num_buf);
			strtrim (resp_buf);

			if ((n = (as_node) calloc (1, AS_NODE_LEN)) IS NULL) {
				fprintf (stderr, "Not enough memory for main node.\n");
				exit (ERROR);
			}

			n->code = atoi (num_buf);	/* set character number */

			if (resp_buf[strlen(resp_buf) - 1] IS '{') { /* dependant list */
				n->str_list = as_make_new_response_list (resp_buf, fp, fname, '}');
			} else if (resp_buf[strlen(resp_buf) - 1] IS '[') { /* non-dep list */
				n->str_list = as_make_new_response_list (resp_buf, fp, fname, ']');
			} else {
				n->str_list = NULL;
			}

			as_sprintf (tmp, resp_buf);

			if ((n->str = (string) calloc (1, strlen (tmp) + 1)) IS NULL) {
				fprintf (stderr, "Not enough memory for response string.\n");
				exit (ERROR);
			}

			strcpy (n->str, tmp);

			if (n3 IS NULL) {
				fprintf (stderr, "Got a response set before a sub-string.\nLine = %s\n", buf);
				exit (ERROR);
			}

			if (n3->str_list IS NULL) {
				n->next = n->prev = NULL;
				n3->str_list = n;
			} else {
				as_add_node (n, n3->str_list);
			}
		} else {	/* new sub-string */
			if (sscanf (buf, "%[^:]", tmp) ISNT 1) {
				fprintf (stderr, "Invalid line in personality table (%s).\nLine = %s\n", fname, buf);
				exit (ERROR);
			}

			strupper (strtrim (tmp));

			if ((n3 = (as_node) calloc (1, AS_NODE_LEN)) IS NULL) {
				fprintf (stderr, "Not enough memory for main node.\n");
				exit (ERROR);
			}

			if ((n3->str = (string) calloc (1, strlen (tmp) + 1)) IS NULL) {
				fprintf (stderr, "Not enough memory for sub-string.\n");
				exit (ERROR);
			}

			strcpy (n3->str, tmp);

			if (n2->str_list IS NULL) {
				n3->next = n3->prev = NULL;
				n2->str_list = n3;
			} else {
				as_add_node (n3, n2->str_list);
			}
		}

		old_pos = ftell (fp);
	}
}

/* as_make_new_response_list - Reads a response list from passed file,
								creating a new node list and returning a
								pointer to it.
*/

as_node as_make_new_response_list (inbuf, fp, fname, end_char)
	string	inbuf;					/* response buffer to maybe append to	*/
	file	fp;						/* file pointer to read list from		*/
	string	fname;					/* name of the file reading from		*/
	char	end_char;				/* character that marks the end of list	*/
{
	char	buf[AS_LINE_LEN+1];		/* line input buffer					*/
	char	tmp[AS_LINE_LEN+1];		/* temp buffer							*/
	as_node	n_hd;					/* head of response list				*/
	as_node	n;						/* node of response list				*/

	n = n_hd = NULL;

	while (fgets (buf, AS_LINE_LEN, fp)) {
		strtrim (buf);

		if (buf[0] IS end_char) {
			strcat (inbuf, buf+1);
			return (n_hd);
		}

		if ((n = (as_node) calloc (1, AS_NODE_LEN)) IS NULL) {
			fprintf (stderr, "Not enough memory for response list node.\n");
			exit (ERROR);
		}

		n->str_list = NULL;

		as_sprintf (tmp, buf);

		if ((n->str = (string) calloc (1, strlen (tmp) + 1)) IS NULL) {
			fprintf (stderr, "Not enough memory for response list string.\n");
			exit (ERROR);
		}

		strcpy (n->str, tmp);

		if (n_hd IS NULL) {
			n->next = n->prev = NULL;
			n_hd = n;
		} else {
			as_add_node (n, n_hd);
		}
	}

	fprintf (stderr, "Unexpected End of File while searching for closing list character (%c).\nFile: '%s'.\n", end_char, fname);
	exit (ERROR);

	return (NULL);	/* shut up the smart-assed compiler */
}


/* as_sprintf - My version of sprintf to handle backslash characters in
				variable string.
*/

void as_sprintf (dest, src)
	string	dest;						/* destination string				*/
	string	src;						/* source string					*/
{
	int		i, j, k;					/* loop counters					*/

	for (i = 0, j = 0; src[i]; ) {
		if (src[i] IS '\\') {
			switch (src[i+1]) {
				case	'f':
					dest[j] = '\f';
					break;
				case	'r':
					dest[j] = '\r';
					break;
				case	'l':
					dest[j] = '\n';
					break;
				case	'n':
					dest[j] = '\n';
					break;
				case	'b':
					dest[j] = '\b';
					break;
				case	't':
					dest[j] = '\t';
					break;
				default:
					if (isdigit (src[i+1])) {
						dest[j] = (char) atoi (&(src[i+1]));
						for (k = 0; isdigit (src[i+k+2]); k++)
							;
						i += k;
					} else {
						dest[j] = src[i+1];
					}
					break;
			}

			i += 2;
			j++;
		} else {
			dest[j++] = src[i++];
		}
	}

	dest[j] = '\0';
}

/* as_char_name - Returns the name of the character/personality with the
					passed number.
*/

string as_char_name (num)
	int		num;					/* the character number					*/
{
	as_str_node	n;					/* temp node pointer					*/
static	char	tmp[MY_BUF_LEN+1];	/* static temp buffer					*/

	n = as_str_node_at_idx (num, AS_Configs.people_names);

	if (n) {
		return (n->str);
	} else {
		sprintf (tmp, "Person %d", num);
		return (tmp);
	}
}

